const { NodeVM } = require('vm2');
const compose = require('koa-compose');

const { getRule } = require('../../util/rule');
const { log } = require('../../util/console');
const createSandbox = require('../../sandbox');
const alphanumericComparator = require('../../util/alphanum');

const ruleMiddleware = async (ctx, next) => {
    // Sort rules in alphanumerical order
    let selectedRules = (ctx.session.selectedRules || []).sort(alphanumericComparator);
    
    let scripts = (await Promise.all(
        selectedRules.map(
            async (name) => ({ name, content: await getRule(name) })
        )
    )).filter(k => k.content != null);

    ctx.session.selectedRules = scripts.map(k => k.name);
    
    let middleware = [];

    for (let script of scripts) {
        let sandbox = new NodeVM({
            console: 'redirect',
            sandbox: createSandbox(ctx, script.name),
            require: {
                external: ['*'],
                builtin: ['*'],
                context: 'sandbox'
            }
        });

        const trace = (e) => {
            if (e.stach && ~e.stack.indexOf('vm.js')) {
                log(ctx, script.name, 'trace', e.stack);
            } else {
                // Errors not generated by rules should be thrown
                throw e;
            }
        }

        const asyncTryCatch = (fun) => {
            // Catch errors during middleware execution
            return async (...args) => {
                try {
                    return await fun(...args);
                } catch (e) {
                    trace(e);
                }
            }
        }

        try {
            sandbox.on('console.log', log.bind(null, ctx, script.name, 'log'));
            sandbox.on('console.error', log.bind(null, ctx, script.name, 'error'));
            sandbox.on('console.trace', trace);
            let mw = sandbox.run(script.content);
            if (Array.isArray(mw)) {
                for (let submw of mw) {
                    middleware.push(asyncTryCatch(mw));
                }
            } else if (mw) {
                middleware.push(asyncTryCatch(mw));
            }
        } catch (e) {
            // Catch errors during module evaluation
            trace(e);
        }
    }

    ctx.rules = selectedRules;
    let composed = compose(middleware);
    await composed(ctx, next);
};

module.exports = ruleMiddleware;